@@ -1,7 +1,7 @@
 /*--------------------------------------------------------------------*\
  * This file is part of Py-notify.                                    *
  *                                                                    *
- * Copyright (C) 2007 Paul Pogonyshev.                                *
+ * Copyright (C) 2007, 2008 Paul Pogonyshev.                          *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public License *
@@ -45,21 +45,19 @@
 
 /* Needed for Py3k, not defined earlier. */
 #ifdef PyVarObject_HEAD_INIT
-#  define Compatibility_VarObject_HEAD_INIT(type, ob_size) PyVarObject_HEAD_INIT (type, ob_size)
+#  define Compatibility_VarObject_HEAD_INIT(ob_size) PyVarObject_HEAD_INIT (0, ob_size)
 #else
-#  define Compatibility_VarObject_HEAD_INIT(type, ob_size) PyObject_HEAD_INIT (type) ob_size,
+#  define Compatibility_VarObject_HEAD_INIT(ob_size) PyObject_HEAD_INIT (0) ob_size,
 #endif
 
 
-/* Type fixup is needed for Python 2.x, but not later (and actually won't work, since
- * structure layout changed a bit).
+/* Type fixup is not really needed for Py3k, but we do this anyway to keep difference with
+ * 2.x version smaller.
  */
 #if defined (PY_MAJOR_VERSION) && PY_MAJOR_VERSION >= 3
-#  define Compatibility_TypeType           &PyType_Type
-#  define Compatibility_Fix_TypeType(type)
+#  define Compatibility_Fix_TypeType(type, meta_type) type.ob_base.ob_base.ob_type = &meta_type;
 #else
-#  define Compatibility_TypeType           NULL
-#  define Compatibility_Fix_TypeType(type) type.ob_type = &PyType_Type;
+#  define Compatibility_Fix_TypeType(type, meta_type) type.ob_type = &meta_type;
 #endif
 
 
@@ -82,6 +80,16 @@
 #endif
 
 
+/* Hide difference between old strings and Unicode strings used in Py3k. */
+#if defined (PY_MAJOR_VERSION) && PY_MAJOR_VERSION >= 3
+#  define Compatibility_String_FromString PyUnicode_FromString
+#  define Compatibility_String_AsString   PyUnicode_AsString
+#else
+#  define Compatibility_String_FromString PyString_FromString
+#  define Compatibility_String_AsString   PyString_AsString
+#endif
+
+
 
 /*- Type forward declarations --------------------------------------*/
 
@@ -110,6 +118,12 @@ typedef  RaisingGCProtector  DebugGCProtector;
 
 /*- Functions forward declarations ---------------------------------*/
 
+static int          GCProtectorMeta_setattro        (PyObject *type,
+                                                     PyObject *name, PyObject *value);
+static PyObject *   GCProtectorMeta_get_default     (PyObject *type, void *context);
+static int          GCProtectorMeta_set_default     (PyObject *type, PyObject *value,
+                                                     void *context);
+
 static void         AbstractGCProtector_dealloc     (PyObject *self);
 static PyObject *   AbstractGCProtector_protect     (PyObject *self,
                                                      PyObject *arguments, PyObject *keywords);
@@ -176,6 +190,21 @@ L{protected <AbstractGCProtector.protect>}.  Of the standard protectors only \
 C{L{RaisingGCProtector}} ever raises these exceptions."
 
 
+#define GC_PROTECTOR_META_DOC "\
+A meta-class for C{L{AbstractGCProtector}}.  Its only purpose is to define C{L{default}} \
+property of the class.  In principle, it can be used for your classes too, but better subclass \
+C{AbstractGCProtector} instead."
+
+#define GC_PROTECTOR_META_DEFAULT_DOC "\
+Current default GC protector.  Starts out as an instance of C{L{FastGCProtector}}, but can be \
+changed for debugging purposes.  Note that setting this class property is only possible if \
+current default protector doesn't have any active protections, i.e. if its \
+C{num_active_protections} property is zero (or has any false truth value).  This is generally \
+true only at the start of the program, so you cannot arbitrarily switch protectors.  Doing so \
+would lead to unpredictable consequences, up to crashing the interpreter, therefore the \
+restriction."
+
+
 #define ABSTRACT_GC_PROTECTOR_DOC "\
 Simple protector interface with two methods for implementations to define."
 
@@ -209,10 +238,7 @@ For convenience, this function always returns C{object} itself.\n\
 #define ABSTRACT_GC_PROTECTOR_SET_DEFAULT_DOC "\
 set_default(protector) \
 \n\
-Set the value of the C{L{default}} variable.  You are advised to do this only once (if at all) \
-somewhere near beginning of your program, because switching protectors being used may even \
-crash Python.  A good reason for calling C{set_default} might be a need to debug protection \
-problems, e.g. with a C{L{DebugGCProtector}}."
+This method is deprecated.  Instead, set C{AbstractGCProtector.default} directly."
 
 
 #define FAST_GC_PROTECTOR_DOC "\
@@ -285,6 +311,56 @@ various protection information.\n\
 
 /*- Types ----------------------------------------------------------*/
 
+PyGetSetDef  GCProtectorMeta_properties[]
+  = { { "default", GCProtectorMeta_get_default, GCProtectorMeta_set_default,
+        GC_PROTECTOR_META_DEFAULT_DOC, NULL },
+      { NULL, NULL, NULL, NULL, NULL } };
+
+PyTypeObject  GCProtectorMeta_Type
+  = { Compatibility_VarObject_HEAD_INIT (0)
+      "notify.gc.GCProtectorMeta",                   /* tp_name           */
+      sizeof (PyHeapTypeObject),                     /* tp_basicsize      */
+      0,                                             /* tp_itemsize       */
+      (destructor)     0,                            /* tp_dealloc        */
+      (printfunc)      0,                            /* tp_print          */
+      (getattrfunc)    0,                            /* tp_getattr        */
+      (setattrfunc)    0,                            /* tp_setattr        */
+      (cmpfunc)        0,                            /* tp_compare        */
+      (reprfunc)       0,                            /* tp_repr           */
+      0,                                             /* tp_as_number      */
+      0,                                             /* tp_as_sequence    */
+      0,                                             /* tp_as_mapping     */
+      (hashfunc)       0,                            /* tp_hash           */
+      (ternaryfunc)    0,                            /* tp_call           */
+      (reprfunc)       0,                            /* tp_str            */
+      (getattrofunc)   0,                            /* tp_getattro       */
+      GCProtectorMeta_setattro,                      /* tp_setattro       */
+      0,                                             /* tp_as_buffer      */
+      Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,      /* tp_flags          */
+      GC_PROTECTOR_META_DOC,                         /* tp_doc            */
+      (traverseproc)   0,                            /* tp_traverse       */
+      (inquiry)        0,                            /* tp_clear          */
+      (richcmpfunc)    0,                            /* tp_richcompare    */
+      0,                                             /* tp_weaklistoffset */
+      (getiterfunc)    0,                            /* tp_iter           */
+      (iternextfunc)   0,                            /* tp_iternext       */
+      0,                                             /* tp_methods        */
+      0,                                             /* tp_members        */
+      GCProtectorMeta_properties,                    /* tp_getset         */
+      0,                                             /* tp_base           */
+      (PyObject *)     0,                            /* tp_dict           */
+      0,                                             /* tp_descr_get      */
+      0,                                             /* tp_descr_set      */
+      0,                                             /* tp_dictoffset     */
+      (initproc)       0,                            /* tp_init           */
+      (allocfunc)      0,                            /* tp_alloc          */
+      (newfunc)        0,                            /* tp_new            */
+      (freefunc)       0,                            /* tp_free           */
+      (inquiry)        0,                            /* tp_is_gc          */
+      (PyObject *)     0,                            /* tp_bases          */
+    };
+
+
 PyMethodDef  AbstractGCProtector_methods[]
   = { { "protect",     (PyCFunction) AbstractGCProtector_protect,
         METH_VARARGS | METH_KEYWORDS,             ABSTRACT_GC_PROTECTOR_PROTECT_DOC },
@@ -295,7 +371,7 @@ PyMethodDef  AbstractGCProtector_methods[]
       { NULL, NULL, 0, NULL } };
 
 PyTypeObject  AbstractGCProtector_Type
-  = { Compatibility_VarObject_HEAD_INIT (Compatibility_TypeType, 0)
+  = { Compatibility_VarObject_HEAD_INIT (0)
       "notify.gc.AbstractGCProtector",               /* tp_name           */
       sizeof (PyObject),                             /* tp_basicsize      */
       0,                                             /* tp_itemsize       */
@@ -352,7 +428,7 @@ PyGetSetDef  FastGCProtector_properties[]
       { NULL, NULL, NULL, NULL, NULL } };
 
 PyTypeObject  FastGCProtector_Type
-  = { Compatibility_VarObject_HEAD_INIT (Compatibility_TypeType, 0)
+  = { Compatibility_VarObject_HEAD_INIT (0)
       "notify.gc.FastGCProtector",                   /* tp_name           */
       sizeof (FastGCProtector),                      /* tp_basicsize      */
       0,                                             /* tp_itemsize       */
@@ -413,7 +489,7 @@ PyGetSetDef  RaisingGCProtector_properties[]
       { NULL, NULL, NULL, NULL, NULL } };
 
 PyTypeObject  RaisingGCProtector_Type
-  = { Compatibility_VarObject_HEAD_INIT (Compatibility_TypeType, 0)
+  = { Compatibility_VarObject_HEAD_INIT (0)
       "notify.gc.RaisingGCProtector",                /* tp_name           */
       sizeof (RaisingGCProtector),                   /* tp_basicsize      */
       0,                                             /* tp_itemsize       */
@@ -463,7 +539,7 @@ PyMethodDef  DebugGCProtector_methods[]
       { NULL, NULL, 0, NULL } };
 
 PyTypeObject  DebugGCProtector_Type
-  = { Compatibility_VarObject_HEAD_INIT (Compatibility_TypeType, 0)
+  = { Compatibility_VarObject_HEAD_INIT (0)
       "notify.gc.DebugGCProtector_Type",             /* tp_name           */
       sizeof (DebugGCProtector),                     /* tp_basicsize      */
       0,                                             /* tp_itemsize       */
@@ -512,11 +588,112 @@ PyTypeObject  DebugGCProtector_Type
 
 static PyObject *  raise_not_implemented_exception = NULL;
 
-static char *      no_keywords[]     = { NULL };
-static char *      object_keywords[] = { "object", NULL };
+static char *      no_keywords[]           = { NULL };
+static char *      object_keywords[]       = { "object", NULL };
 
 static PyObject *  unprotection_error_type = NULL;
 
+static PyObject *  default_protector       = NULL;
+static PyObject *  default_attribute_name  = NULL;
+
+
+
+/*- GCProtectorMeta type methods -----------------------------------*/
+
+static int
+GCProtectorMeta_setattro (PyObject *type, PyObject *name, PyObject *value)
+{
+  switch (PyObject_RichCompareBool (name, default_attribute_name, Py_EQ))
+    {
+    case 1:
+      return PyObject_GenericSetAttr (type, name, value);
+
+    case 0:
+      return PyType_Type.tp_setattro (type, name, value);
+
+    default:
+      return -1;
+    }
+}
+
+
+static PyObject *
+GCProtectorMeta_get_default (PyObject *type, void *context)
+{
+  Py_INCREF (default_protector);
+  return default_protector;
+}
+
+static int
+GCProtectorMeta_set_default (PyObject *type, PyObject *value, void *context)
+{
+  if (value == default_protector)
+    return 0;
+
+  switch (PyObject_IsInstance (value, (PyObject *) &AbstractGCProtector_Type))
+    {
+    case 1:
+      {
+        PyObject *num_active_protections = PyObject_GetAttrString (default_protector,
+                                                                   "num_active_protections");
+
+        if (num_active_protections)
+          {
+            switch (PyObject_IsTrue (num_active_protections))
+              {
+              case 0:
+                Py_DECREF (num_active_protections);
+                break;
+
+              case 1:
+                {
+                  PyObject *as_string = PyObject_Str (num_active_protections);
+
+                  if (as_string)
+                    {
+                      PyErr_Format (PyExc_ValueError,
+                                    ("cannot set a different GC protector: current has active "
+                                     "protections (num_active_protections = %.200s)"),
+                                    Compatibility_String_AsString (as_string));
+                    }
+                  else
+                    {
+                      PyErr_Clear ();
+                      PyErr_SetString (PyExc_ValueError, ("cannot set a different GC protector: "
+                                                          "current has active protections"));
+                    }
+                }
+
+              default:
+                Py_DECREF (num_active_protections);
+                return -1;
+              }
+          }
+        else
+          {
+            /* Assume that there is no such attribute then. */
+            PyErr_Clear ();
+          }
+
+        Py_DECREF (default_protector);
+
+        default_protector = value;
+        Py_INCREF (default_protector);
+
+        return 0;
+      }
+
+    case 0:
+      PyErr_Format (PyExc_ValueError,
+                    ("can only set AbstractGCProtector.default to an instance of "
+                     "AbstractGCProtector; got %.200s instead"),
+                    value->ob_type->tp_name);
+      break;
+    }
+
+  return -1;
+}
+
 
 
 /*- AbstractGCProtector type methods -------------------------------*/
@@ -556,6 +733,9 @@ AbstractGCProtector_unprotect (PyObject *self, PyObject *arguments, PyObject *ke
 }
 
 
+/* Deprecated old function.  Assigning directly to `default' attribute is now possible and
+ * is preferred.
+ */
 static PyObject *
 AbstractGCProtector_set_default (PyObject *null, PyObject *arguments, PyObject *keywords)
 {
@@ -568,7 +748,7 @@ AbstractGCProtector_set_default (PyObject *null, PyObject *arguments, PyObject *
                                     protector_keywords, &AbstractGCProtector_Type, &new_protector))
     return NULL;
 
-  if (PyDict_SetItemString (AbstractGCProtector_Type.tp_dict, "default", new_protector) == -1)
+  if (GCProtectorMeta_set_default (NULL, new_protector, NULL) == -1)
     return NULL;
 
   Py_INCREF (Py_None);
@@ -881,10 +1061,10 @@ DebugGCProtector_unprotect (DebugGCProtector *self, PyObject *arguments, PyObjec
 
 /*- Module initialization ------------------------------------------*/
 
-#define REGISTER_TYPE(dictionary, type, name, error_label)              \
+#define REGISTER_TYPE(dictionary, type, meta_type, name, error_label)   \
   do                                                                    \
     {                                                                   \
-      Compatibility_Fix_TypeType (type);                                \
+      Compatibility_Fix_TypeType (type, meta_type);                     \
       type.tp_alloc = PyType_GenericAlloc;                              \
       type.tp_new   = PyType_GenericNew;                                \
       if (PyType_Ready (&type) == -1                                    \
@@ -903,7 +1083,6 @@ initgc (void)
   PyObject *dictionary;
   PyObject *utilities                     = NULL;
   PyObject *unprotection_error_dictionary = NULL;
-  PyObject *default_protector             = NULL;
 
   module = Py_InitModule ("notify.gc", NULL);
   if (!module)
@@ -913,6 +1092,10 @@ initgc (void)
   if (!dictionary)
     goto error;
 
+  default_attribute_name = Compatibility_String_FromString ("default");
+  if (!default_attribute_name)
+    goto error;
+
   utilities = PyImport_ImportModule ("notify.utils");
   if (!utilities)
     goto error;
@@ -945,20 +1128,20 @@ initgc (void)
   if (PyDict_SetItemString (dictionary, "UnprotectionError", unprotection_error_type) == -1)
     goto error;
 
-  REGISTER_TYPE (dictionary, AbstractGCProtector_Type, "AbstractGCProtector", error);
-  REGISTER_TYPE (dictionary, FastGCProtector_Type,     "FastGCProtector",     error);
-  REGISTER_TYPE (dictionary, RaisingGCProtector_Type,  "RaisingGCProtector",  error);
-  REGISTER_TYPE (dictionary, DebugGCProtector_Type,    "DebugGCProtector",    error);
+  GCProtectorMeta_Type.tp_base = &PyType_Type;
+
+  REGISTER_TYPE (dictionary, GCProtectorMeta_Type,     PyType_Type, "GCProtectorMeta",     error);
+  REGISTER_TYPE (dictionary, AbstractGCProtector_Type, GCProtectorMeta_Type,
+                 "AbstractGCProtector", error);
+
+  REGISTER_TYPE (dictionary, FastGCProtector_Type,     PyType_Type, "FastGCProtector",     error);
+  REGISTER_TYPE (dictionary, RaisingGCProtector_Type,  PyType_Type, "RaisingGCProtector",  error);
+  REGISTER_TYPE (dictionary, DebugGCProtector_Type,    PyType_Type, "DebugGCProtector",    error);
 
   default_protector = FastGCProtector_new ();
   if (!default_protector)
     goto error;
 
-  if (PyDict_SetItemString (AbstractGCProtector_Type.tp_dict, "default", default_protector) == -1)
-    goto error;
-
-  Py_DECREF (default_protector);
-
   if (PyModule_AddStringConstant (module, "__doc__", MODULE_DOC) == -1)
     goto error;
   if (PyModule_AddStringConstant (module, "__docformat__", "epytext en") == -1)
