@@ -3,7 +3,7 @@
 #--------------------------------------------------------------------#
 # This file is part of Py-notify.                                    #
 #                                                                    #
-# Copyright (C) 2007 Paul Pogonyshev.                                #
+# Copyright (C) 2007, 2008 Paul Pogonyshev.                          #
 #                                                                    #
 # This library is free software; you can redistribute it and/or      #
 # modify it under the terms of the GNU Lesser General Public License #
@@ -25,8 +25,44 @@
 import gc
 import unittest
 
+from notify.gc import AbstractGCProtector
 
-__all__ = ('NotifyTestCase', 'ignoring_exceptions')
+
+__all__ = ('NotifyTestObject', 'NotifyTestCase', 'ignoring_exceptions')
+
+
+
+# Each test should create an instance of this class and connect handlers defined here.
+# The reason is that conditions, variables etc. are garbage-collected only after last
+# handler dies.  If handlers belong to an object which is in test function scope,
+# everything used by the test will be GC-collected after test finishes.  Else we can have
+# dangling conditions, which will prevent us from changing `AbstractGCProtector.default',
+# among other things.
+class NotifyTestObject (object):
+
+    def __init__(self):
+        self.results = []
+
+
+    def assert_results (self, *results):
+        valid_results = list (results)
+
+        if self.results != valid_results:
+            raise AssertionError ('results: %s; expected: %s' % (self.results, valid_results))
+
+
+    def simple_handler (self, *arguments):
+        if len (arguments) == 1:
+            arguments = arguments[0]
+
+        self.results.append (arguments)
+
+    def simple_handler_100 (self, *arguments):
+        self.simple_handler (100 + arguments[0])
+
+
+    def simple_handler_200 (self, *arguments):
+        self.simple_handler (200 + arguments[0])
 
 
 
@@ -36,9 +72,25 @@ class NotifyTestCase (unittest.TestCase):
 
 
     def setUp (self):
-        gc.set_threshold (0, 0, 0)
         super (NotifyTestCase, self).setUp ()
 
+        gc.set_threshold (0, 0, 0)
+        self.__num_active_protections = AbstractGCProtector.default.num_active_protections
+
+
+    # It is important to leave no garbage behind, since `AbstractGCProtector.default' is
+    # only assignable in 'fresh' state.
+    def tearDown (self):
+        super (NotifyTestCase, self).tearDown ()
+
+        self.collect_garbage ()
+
+        if AbstractGCProtector.default.num_active_protections != self.__num_active_protections:
+            raise ValueError (('number of active GC protections before and after the test differ: '
+                               '%d != %d')
+                              % (self.__num_active_protections,
+                                 AbstractGCProtector.default.num_active_protections))
+
 
     def non_existing_attribute_setter (self, object, name = 'this_attribute_sure_doesnt_exist'):
         return lambda: setattr (object, name, None)
@@ -67,30 +119,36 @@ class NotifyTestCase (unittest.TestCase):
         # Note: hashes are _not_ required to be different, so don't test them.
 
 
-    def assert_results (self, *results):
-        valid_results = list (results)
+    def collect_garbage (self, times = None):
+        if times:
+            # This is the old way, not used now.  We use a slower, but more robust way of
+            # 'collect while collects' below.
+            for k in range (times):
+                gc.collect ()
+        else:
+            num_objects = self.__count_garbage_collectable_objects ()
 
-        if self.results != valid_results:
-            raise AssertionError ('results: %s; expected: %s' % (self.results, valid_results))
+            # TODO: Account for 'statics' like notify.Condition.TRUE.  Condition below is
+            #       always false due to 'statics'.
+            if num_objects == 0:
+                return;
 
+            num_passes  = 0
+            while num_passes < 20:
+                gc.collect ()
+                num_objects_new = self.__count_garbage_collectable_objects ()
 
-    def simple_handler (self, *arguments):
-        if len (arguments) == 1:
-            arguments = arguments[0]
+                if num_objects_new < num_objects:
+                    num_objects  = num_objects_new
+                    num_passes  += 1
+                else:
+                    break
 
-        self.results.append (arguments)
+    def __count_garbage_collectable_objects ():
+        return len ([object for object in gc.get_objects ()
+                     if type (object).__module__.startswith ('notify.')])
 
-    def simple_handler_100 (self, *arguments):
-        self.simple_handler (100 + arguments[0])
-
-
-    def simple_handler_200 (self, *arguments):
-        self.simple_handler (200 + arguments[0])
-
-
-    def collect_garbage (self, times = 1):
-        for k in range (0, times):
-            gc.collect ()
+    __count_garbage_collectable_objects = staticmethod (__count_garbage_collectable_objects)
 
 
     def note_skipped_tests (tests_defined = False):
